#!/usr/bin/env python
import numpy as np
import pandas as pd
from tools import data_loader
from pathlib import Path
from sklearn.base import BaseEstimator, TransformerMixin
from scipy.stats import skew, boxcox
from sklearn.preprocessing import StandardScaler
from sklearn.preprocessing import StandardScaler, MinMaxScaler
from sklearn.preprocessing import RobustScaler, MaxAbsScaler, QuantileTransformer
from tools.classifiers import KnnClassifier, SvmClassifier, LrClassifier
from tools.classifiers import XgbClassifier, RfClassifier, MlpClassifier
from utility.settings import load_settings
import csv
from typing import List
import paths as pt
from utility.metrics import compute_mean, compute_std

class DummyScaler(BaseEstimator, TransformerMixin):
    def fit_transform(self, X):
        return np.array(X)

class DummyNormalizer(BaseEstimator, TransformerMixin): 
    def fit_transform(self, X, case=None):
        return np.array(X)

class BoxCoxNormalizer(BaseEstimator, TransformerMixin):
    def fit_transform(self, X, case=None):
        if case == "Risk":
            numeric_feats = ['Gender_Male', 'Gender_Female', 'BirthYear', 'Cluster',
                    'LoanPeriod', 'NumberSplit', 'NumberScreening', 'NumberWeeks',
                    'MeanEvaluation', 'NumberFalls', 'NumberTraining', 'NumberTrainingWeek',
                    'TimeBetweenTraining', 'NumberWeeksNoTraining', 'Needs', 'Physics',
                    'NumberAts', 'NumberEx']
        else:
            numeric_feats = ['Gender_Male', 'Gender_Female', 'BirthYear',
                            'Cluster', 'LoanPeriod', 'NumberAts']
        skewed_feats = X[numeric_feats].apply(lambda x: skew(x.dropna()))
        skewed_feats = skewed_feats[skewed_feats > 0.25]
        skewed_feats = skewed_feats.index
        for feats in skewed_feats:
            X[feats] = X[feats] + 1
            X[feats], lam = boxcox(X[feats])
        return np.array(X)
    
class BoxCoxNormalizerNoGender(BaseEstimator, TransformerMixin):
    def fit_transform(self, X, case=None):
        if case == "Risk":
            numeric_feats = ['BirthYear', 'Cluster',
                    'LoanPeriod', 'NumberSplit', 'NumberScreening', 'NumberWeeks',
                    'MeanEvaluation', 'NumberFalls', 'NumberTraining', 'NumberTrainingWeek',
                    'TimeBetweenTraining', 'NumberWeeksNoTraining', 'Needs', 'Physics',
                    'NumberAts', 'NumberEx']
        else:
            numeric_feats = ['BirthYear', 'Cluster', 'LoanPeriod', 'NumberAts']
        skewed_feats = X[numeric_feats].apply(lambda x: skew(x.dropna()))
        skewed_feats = skewed_feats[skewed_feats > 0.25]
        skewed_feats = skewed_feats.index
        for feats in skewed_feats:
            X[feats] = X[feats] + 1
            X[feats], lam = boxcox(X[feats])
        return np.array(X)

def main():
    cases = ["Complete", "Compliance", "Fall", "Risk"]
    metrics = ['accuracy', 'precision', 'recall', 'roc_auc', 'average_precision', 'f1']
    normalizer_names = ["None", "BoxCox", "BoxCoxNoGender"]
    normalizers = [DummyNormalizer(), BoxCoxNormalizer(), BoxCoxNormalizerNoGender()]
    scaler_names = ["None", "Standard", "MinMax", "MinMaxRange", "Robust",
    "MaxAbs", "QuantileTransformer", "QuantileTransformerNorm"]
    scalers = [DummyScaler(), StandardScaler(), MinMaxScaler(), MinMaxScaler((-1, 1)), RobustScaler(),
            MaxAbsScaler(), QuantileTransformer(), QuantileTransformer(random_state=0),
            QuantileTransformer(output_distribution='normal', random_state=0)]
    for case in cases:
        output_filename = f"{case} model preprocessing.csv"
        header = ['clf', 'normalizer', 'scaler', 'accuracy_mean', 'accuracy_std',
                  'precision_mean', 'precision_std', 'recall_mean', 'recall_std',
                  'roc_auc_mean', 'roc_auc_std', 'pr_auc_mean', 'pr_auc_std',
                  'f1_mean', 'f1_std']
        with open(Path.joinpath(pt.REPORTS_DIR, output_filename), 'w',
                  encoding='UTF8', newline='') as f:
            writer = csv.writer(f)
            writer.writerow(header)
        for normalizer_name, normalizer in zip(normalizer_names, normalizers):
            for scaler_name, scaler in zip(scaler_names, scalers):
                if case == "Complete":
                    settings = load_settings("complete_emb.yaml")
                    dl = data_loader.CompleteDataLoader("complete_emb.csv", settings).load_data()
                    X, y = dl.get_data()
                elif case == "Compliance":
                    settings = load_settings("compliance_emb.yaml")
                    dl = data_loader.ComplianceDataLoader("compliance_emb.csv", settings).load_data()
                    X, y = dl.get_data()
                elif case == "Fall":
                    settings = load_settings("fall_emb.yaml")
                    dl = data_loader.FallDataLoader("fall_emb.csv", settings).load_data()
                    X, y = dl.get_data()
                else:
                    settings = load_settings("risk_emb.yaml")
                    dl = data_loader.RiskDataLoader("risk_emb.csv", settings).load_data()
                    X, y = dl.get_data()
                    
                emb_cols = X.filter(regex='((\d+)[Ats|Ex])\w+', axis=1)
                n_norm_cols = X.shape[1] - emb_cols.shape[1]
                X_sc = pd.DataFrame(normalizer.fit_transform(X.iloc[:,:n_norm_cols], case))
                X = pd.concat([X_sc, X.iloc[:,n_norm_cols:]], axis=1)
                X = np.array(X)
                y = np.array(y)
                X_sc = scaler.fit_transform(X[:,:n_norm_cols])
                X = np.concatenate([X_sc, X[:,n_norm_cols:]], axis=1)
                results = {'KNN': KnnClassifier(X, y).evaluate(metrics),
                           'SVM': SvmClassifier(X, y).evaluate(metrics),
                           'LR': LrClassifier(X, y).evaluate(metrics),
                           'XGB': XgbClassifier(X, y).evaluate(metrics),
                           'RF': RfClassifier(X, y).evaluate(metrics),
                           'MLP': MlpClassifier(X, y).evaluate(metrics)}
                
                with open(Path.joinpath(pt.REPORTS_DIR, output_filename), 'a',
                          encoding='UTF8', newline='') as f:
                    writer = csv.writer(f)
                    for clf_name, result in results.items():
                        data = [clf_name, normalizer_name, scaler_name]
                        for metric in metrics:
                            mean = compute_mean(result[metric])
                            std = compute_std(result[metric])
                            data.extend((mean, std))
                        writer.writerow(data)

if __name__ == '__main__':
    main()