#!/usr/bin/env python
from pathlib import Path
import yaml
import numpy as np
import pandas as pd
import paths as pt
from tools import file_reader, file_writer, feature_maker
from utility import data_dto, dataset
from pandas.tseries.offsets import DateOffset

def main():
    sc = file_reader.read_pickle(pt.INTERIM_DATA_DIR, 'sc.pkl')
    ss = file_reader.read_pickle(pt.INTERIM_DATA_DIR, 'ss.pkl')
    td = file_reader.read_pickle(pt.INTERIM_DATA_DIR, 'td.pkl')
    tc = file_reader.read_pickle(pt.INTERIM_DATA_DIR, 'tc.pkl')
    ats = file_reader.read_pickle(pt.INTERIM_DATA_DIR, 'ats.pkl')
    fd = file_reader.read_pickle(pt.INTERIM_DATA_DIR, 'fd.pkl')
    
    with open(Path.joinpath(pt.CONFIGS_DIR, "data.yaml"), 'r') as stream:
        settings = yaml.safe_load(stream)
    
    data = data_dto.Data(sc, ss, td, tc, ats, fd)
    screenings = get_screenings(data, settings)
    file_writer.write_csv(screenings, pt.INTERIM_DATA_DIR, 'screenings.csv')

def get_screenings(data, settings):
    ids = dataset.create_union_of_ids(data.sc, data.ss, data.td, data.tc)
    all_screenings = pd.DataFrame()
    for id in ids:
        screenings = get_screenings_by_id(data, id, settings)
        all_screenings = pd.concat([all_screenings, screenings], axis=0, ignore_index=True)
    return all_screenings

def get_screenings_by_id(data, id, settings):
    citizen_data = feature_maker.get_citizen_data(data, id)
    
    if citizen_data.sc.empty:
        return
    
    # Remove duplicate falls
    citizen_data.fd = citizen_data.fd.drop_duplicates(["CitizenId", "Date"])
    
    # Record split dates
    first_screening_date = citizen_data.sc.ScreeningDate.iloc[0]
    split_dates = []
    for i, date in enumerate(list(citizen_data.sc.ScreeningDate)):
        if date > (first_screening_date + DateOffset(weeks=15)):
            split_dates.append(citizen_data.sc.ScreeningDate.iloc[i-1])
            first_screening_date = date
    
    # Split screenings
    split_screenings = []
    split_dates.append(citizen_data.sc.ScreeningDate.iloc[-1]) #add boundary
    for split_date in split_dates:
        df = citizen_data.sc.loc[citizen_data.sc['ScreeningDate'] <= split_date]
        split_screenings.append(df)
        citizen_data.sc = citizen_data.sc.loc[citizen_data.sc['ScreeningDate'] > split_date]
    
    screenings = pd.DataFrame()
    for n_split, screening_split in enumerate(split_screenings):
        for n_screening, screening in enumerate(screening_split.itertuples(index=False)):
            single_screening = pd.DataFrame([{'CitizenId': id}])
            single_screening['Gender'] = feature_maker.get_gender(screening_split)
            single_screening['BirthYear'] = feature_maker.get_birth_year(screening_split)
            single_screening['NumberSplit'] = n_split
            single_screening['NumberScreening'] = n_screening

            # Get time dependent window data
            if n_screening == 0:
                start_date = screening_split.iloc[n_screening].ScreeningDate
            else:
                start_date = screening_split.iloc[n_screening-1].ScreeningDate
            end_date = screening.ScreeningDate
            tds, tcs, _ = feature_maker.get_screening_data(citizen_data.td,
                                                           citizen_data.tc,
                                                           citizen_data.ss,
                                                           start_date,
                                                           end_date)
                        
            # Record number of trainings and falls after screening end
            n_weeks_after = 0
            n_training_after = 0
            n_falls_after = 0
            if (n_screening == len(screening_split) - 1):
                screening_date = screening.ScreeningDate
                trainings_after = citizen_data.td.loc[(citizen_data.td['RatingDate'] > screening_date)]
                fall_end_date = screening_date + DateOffset(years=1)
                trainings_end_date = screening_date + DateOffset(weeks=15)
                n_falls_after = len(citizen_data.fd.loc[(citizen_data.fd['Date'] > screening_date) &
                                                        (citizen_data.fd['Date'] <= fall_end_date)])
                trainings_before_15 = trainings_after.loc[(trainings_after['RatingDate'] <= trainings_end_date)]
                if len(trainings_before_15) > 0:
                    day_start = trainings_before_15.RatingDate.iloc[0]
                    day_end = trainings_before_15.RatingDate.iloc[-1]
                    n_weeks_after = (day_end - day_start).days / 7
                    n_training_after = len(trainings_before_15)

            # Overwrite start date with rating date if it exists
            start_date = tds.RatingDate.iloc[0] if not tds.empty else start_date
            
            date_format = '%d-%m-%Y'
            start_date = feature_maker.convert_date_to_datetime(start_date, date_format)
            end_date = feature_maker.convert_date_to_datetime(end_date, date_format)
            single_screening['StartDate'] = start_date
            single_screening['EndDate'] = end_date

            n_weeks = feature_maker.get_interval_length(start_date, end_date)
            single_screening['NumberWeeks'] = np.around(n_weeks + n_weeks_after, decimals=2)
            single_screening['MeanEvaluation'] = feature_maker.get_mean_evaluation(tds)
            single_screening['StdEvaluation'] = feature_maker.get_std_evaluation(tds)
            single_screening['MinEvaluation'] = feature_maker.get_min_evaluation(tds)
            single_screening['MaxEvaluation'] = feature_maker.get_max_evaluation(tds)

            n_training_window = feature_maker.get_n_training_window(tds)
            n_falls = feature_maker.get_number_falls(citizen_data.fd, start_date, end_date)
            single_screening['NumberFalls'] = n_falls + n_falls_after
            if screening.ScreeningDate in list(tds.RatingDate):
                single_screening['NumberTraining'] = n_training_window + n_training_after
            else:
                single_screening['NumberTraining'] = n_training_window + n_training_after + 1
            single_screening['NumberTrainingWeek'] = feature_maker.get_n_training_week(n_weeks, n_training_window)

            training_pr_week = feature_maker.get_training_week(tds, start_date)
            n_weeks_with_training = feature_maker.get_n_weeks_with_training(tds, start_date)
            single_screening['MaxNumberTrainingWeek'] = feature_maker.get_n_training_week_max(training_pr_week)
            single_screening['MinNumberTrainingWeek'] = feature_maker.get_n_training_week_min(training_pr_week,
            n_weeks_with_training, n_weeks)
            single_screening['NumberWeeksWithTraining'] = n_weeks_with_training
            single_screening['TimeBetweenTraining'] = feature_maker.get_time_between_training_mean(tds)
            single_screening['NumberWeeksNoTraining'] = feature_maker.get_n_weeks_without_training(n_weeks,
            n_weeks_with_training)
            single_screening['NumberCancels'] = tcs.shape[0]
            single_screening['TimeBetweenCancels'] = feature_maker.get_mean_time_between_cancels(tcs)
            single_screening['NumberCancelsWeek'] = feature_maker.get_mean_cancels_week(tcs.shape[0], n_weeks)
            cancels_week = feature_maker.get_cancels_week(tcs)
            single_screening['MinNumberCancelsWeek'] = feature_maker.get_n_cancel_week_min(cancels_week)
            single_screening['MaxNumberCancelsWeek'] = feature_maker.get_n_training_week_max(cancels_week)

            single_screening['Ats'] = feature_maker.get_ats(citizen_data.ats, end_date, settings)
            single_screening['NumberAts'] = feature_maker.get_number_ats(citizen_data.ats, end_date)
            single_screening['LoanPeriod'] = feature_maker.get_avg_loan_period(citizen_data.ats, end_date)

            single_screening['Needs'] = screening.NeedForHelpScore
            single_screening['NeedsReason'] = screening.NeedForHelpReason
            single_screening['Physics']= screening.PhysicalStrengthScore
            single_screening['PhysicsReason']= screening.PhysicalStrengthReason

            single_screening['Ex'] = screening.ExerciseContent
            single_screening['NumberEx'] = feature_maker.get_number_exercises(screening.ExerciseContent)
            single_screening['HasFallRisk'] = sum(map(
                screening.ExerciseContent.count,
                settings['fall_exercises'])) > settings['fall_exercise_threshold']

            screenings = pd.concat([screenings, single_screening], axis=0, ignore_index=True)
        
    return screenings

if __name__ == "__main__":
    main()