# %%
from pathlib import Path
from tools import file_reader
from sklearn.model_selection import train_test_split
from tensorflow.keras.utils import to_categorical
import tensorflow as tf
import numpy as np
import pickle

path = Path('C:/Users/Daniel/Documents/air/R_and_D_Daniel_Thomas') # Local path for data-files (ats.csv etc.)

#%%
# Get data
def load_date(path):
    folder = 'raw_data'
    raw_df = file_reader.read_pickle(path, f'{folder}/ats.pkl').reset_index(drop=True)
    mapping = file_reader.read_csv(path, f'{folder}/ats.csv',
                                converters={f'{folder}/ats_id': str})
    raw_df = raw_df.loc[raw_df['Gender'] != 'nan']
    print('Loaded data')
    return raw_df, mapping

# Shorten iso class
def shorten_iso(raw_df, mapping): 
    df = raw_df.copy() 
    df['DevISOClass'] = df['DevISOClass'].apply(lambda x: x[:6])
    df = df.dropna(subset=['CitizenId'])
    mapping_dict = dict(mapping.values)
    df = df.replace(to_replace=mapping_dict)
    df.sort_values(by='LendDate', inplace=True)
    cond = df['DevISOClass'].str.isdigit()
    df.drop(df[~cond].index, inplace=True)
    print('ISO-numbers shortened')

    return df

# Remove citizens with only 1 record
def remove_single_records(df, seq_length):
    counts = df.CitizenId.value_counts()
    df = df[~df['CitizenId'].isin(counts[counts < seq_length + 1].index)]
    print('Single records removed')

    return df

# Create list of sequence
def create_sequence_list_LSTM(df, seq_length, column_name='DevISOClass'):
    df = remove_single_records(df, seq_length)
    X = []
    y = []
    print(f'shape: {df.shape}')
    for citizenId in df.CitizenId.unique():
        temp = df.loc[df.CitizenId==citizenId]
        seq = tf.keras.preprocessing.sequence.TimeseriesGenerator(temp[column_name].values, 
                                                                temp[column_name].values, 
                                                                length=seq_length,
                                                                batch_size=1,
                                                                shuffle=False)
        
        [((X.append(x[0][0].tolist())), y.append(x[1])) for x in seq]

    return X, y

# Create list of sequence with zero-padding
def create_sequence_list_LSTM_pad(df, seq_length, column_name='DevISOClass'):
    X = []
    y = []
    for temp_seq in range(seq_length):
        temp_seq = temp_seq + 1
        temp_df = remove_single_records(df, temp_seq)

        for i, citizenId in enumerate(temp_df.CitizenId.unique()):            
            #  Below if is to not create a sequence when it's the first sample
            if i == 0:
                continue

            temp = temp_df.loc[temp_df.CitizenId==citizenId]
            seq = tf.keras.preprocessing.sequence.TimeseriesGenerator(temp[column_name].values, 
                                                                    temp[column_name].values, 
                                                                    length=temp_seq,
                                                                    batch_size=1,
                                                                    shuffle=False)
            for x in seq:
                to_append = np.pad(x[0][0], (seq_length-temp_seq, 0), 'constant', constant_values=(0))
                X.append(to_append)
                y.append(x[1])
    return X, y

def make_strings_categorical(string_list, mapping_to_numbers={}):
    y_true = np.zeros((len(string_list)))
    for i, raw_label in enumerate(string_list):
        if raw_label not in mapping_to_numbers:
            mapping_to_numbers[raw_label] = len(mapping_to_numbers)
        y_true[i] = mapping_to_numbers[raw_label] +1
    
    mapping_to_numbers['ZeroPadding'] = 0
    return mapping_to_numbers

def convert_str_to_int(word, mapping):
    return mapping[word]

def load_pickle(seq_length):
    with open(f'preprocessed_data_LSTM/X_{seq_length}.pkl', 'rb') as f:
        X = pickle.load(f)
    with open(f'preprocessed_data_LSTM/y_{seq_length}.pkl', 'rb') as f:
        y = pickle.load(f)
    return X, y


#%%
# Initialize settings for run
LOAD_PREPROCESSED_DATA = False
PREPROCESSED_SEQUENCE_NAME = 10
ZERO_PAD_SEQUENCES = True
RUN_SMALLER_SAMPLE_SIZE = False
seq_length = 10
OVERWRITE_LOCAL_PREPROCESSED_SEQUENCES = False

#%%

raw_df, mapping = load_date(path)

if RUN_SMALLER_SAMPLE_SIZE:
    raw_df = raw_df.head(50000)
    counts = raw_df.DevISOClass.value_counts()
    # raw_df = raw_df.loc[raw_df.DevISOClass.isin(counts.index[counts > seq_length])]

df = shorten_iso(raw_df, mapping)
#%%
string_to_int_mapping = make_strings_categorical(df.DevISOClass.values)
vocabulary_size = len(string_to_int_mapping)+1
df['DevISOClassCategorical'] = np.vectorize(convert_str_to_int)(df['DevISOClass'], string_to_int_mapping)

# Create sequences
# Load already processed sequences to save time 
if LOAD_PREPROCESSED_DATA:
    X, y = load_pickle(PREPROCESSED_SEQUENCE_NAME)

elif ZERO_PAD_SEQUENCES:
    X, y = create_sequence_list_LSTM_pad(
        df[['DevISOClassCategorical', 'CitizenId']], 
        seq_length, 
        column_name='DevISOClassCategorical')

else:
    X, y = create_sequence_list_LSTM(
        df, 
        seq_length, 
        column_name='DevISOClassCategorical')
# %%
# Store sequences locally in pickle-file
if OVERWRITE_LOCAL_PREPROCESSED_SEQUENCES:
    with open(f'preprocessed_data_LSTM/X_{seq_length}.pkl', 'wb') as f:
        pickle.dump(X, f)
    with open(f'preprocessed_data_LSTM/y_{seq_length}.pkl', 'wb') as f:
        pickle.dump(y, f)
# %%
X_train, X_test, y_train, y_test = train_test_split(X, y, train_size=0.8)
#%%
train_targets = to_categorical(y_train, num_classes=vocabulary_size)
test_targets = to_categorical(y_test, num_classes=vocabulary_size)

# %%
def topKacc_5(Y_true, Y_pred):
  return top_k_categorical_accuracy(Y_true, 
                                    Y_pred, 
                                    k = 5)
def topKacc_4(Y_true, Y_pred):
  return top_k_categorical_accuracy(Y_true, 
                                    Y_pred,
                                    k = 4)
def topKacc_3(Y_true, Y_pred):
  return top_k_categorical_accuracy(Y_true, 
                                    Y_pred, 
                                    k = 3)

def topKacc_2(Y_true, Y_pred):
  return top_k_categorical_accuracy(Y_true, 
                                    Y_pred, 
                                    k = 2)

#%%
%load_ext tensorboard

# from tcn import TCN, tcn_full_summary
from keras.callbacks import EarlyStopping
from keras.models import Sequential
from keras.layers import LSTM, Dense, Embedding, Dropout, CuDNNLSTM
from keras.metrics import top_k_categorical_accuracy
import datetime

filepath = './models/' + datetime.datetime.now().strftime("%Y%m%d-%H%M") + '/{epoch:02d}-{val_loss:.2f}.hdf5'
log_dir = "logs/fit/" + datetime.datetime.now().strftime("%Y%m%d-%H%M%S")

tensorboard_callback = tf.keras.callbacks.TensorBoard(log_dir=log_dir, histogram_freq=1)
checkpoint = tf.keras.callbacks.ModelCheckpoint(filepath=filepath, mode='min', monitor='val_loss', verbose=2, save_best_only=True)
es = EarlyStopping(monitor='val_loss', patience=5)
callbacks_list = [checkpoint, es, tensorboard_callback]

optimizer = tf.keras.optimizers.Adam(learning_rate=0.002)

model = Sequential()
model.add(Embedding(vocabulary_size, 300, input_length=seq_length))
model.add(CuDNNLSTM(units=1024, input_shape=(1,seq_length)))
model.add(Dense(vocabulary_size, activation='softmax'))
model.compile(optimizer=optimizer, loss='categorical_crossentropy', metrics=['categorical_accuracy', topKacc_2, topKacc_3, topKacc_4, topKacc_5])
model.summary()

num_epochs = 500
history = model.fit(
    np.array(X_train),
    train_targets,
    validation_data=(np.array(X_test), test_targets),
    epochs=num_epochs, 
    verbose=1, 
    use_multiprocessing=True,
    callbacks=callbacks_list
)

#%%
%tensorboard --logdir logs/fit
# history.history['val_accuracy']
