import pandas as pd
import numpy as np
import os
from abc import ABC, abstractmethod
from typing import List

class BaseCleaner(ABC):
    @abstractmethod
    def clean_clusters(self, cl):
        """Cleans the cluster data set"""
    
    @abstractmethod
    def clean_patient_data(self, ptd):
        """Cleans the patient data set"""
    
    @abstractmethod
    def clean_screening_content(self, sc, ptd):
        """Cleans the screening content data set"""
        
    @abstractmethod
    def clean_status_set(self, ss, ptd):
        """Cleans the status set data set"""
    
    @abstractmethod
    def clean_training_done(self, td, ptd):
        """Cleans the training done data set"""
        
    @abstractmethod
    def clean_training_cancelled(self, tc, ptd):
        """Cleans the training cancelled data set"""    

    @abstractmethod
    def clean_assistive_aids(self, ats, ic, ids):
        """Cleans the assistive aids data set"""
        
    @abstractmethod
    def clean_fall_data(self, fd):
        """Cleans the fall set"""
        
    def remove_citizens_not_in_patient_data(self, train_data: pd.DataFrame,
                                            patient_data: pd.DataFrame,
                                            id: str) -> pd.DataFrame:
        data = train_data[train_data[id].isin(patient_data[id].unique())]
        return data

    def remove_citizens_without_valid_id(self, df: pd.DataFrame) -> pd.DataFrame:
        df = df[df['CitizenId'] != "0000000000"]
        df = df[df['CitizenId'] != '0']
        df = df[df['CitizenId'] != "#VALUE!"]
        df = df[df['CitizenId'] != '681']
        df = df.dropna(subset=['CitizenId'])
        return df

    def merge_train_and_patient_data(self, train_data: pd.DataFrame,
                                    patient_data: pd.DataFrame,
                                    key: str) -> pd.DataFrame:
        return pd.merge(train_data, patient_data, on=key)

    def sort_dataframe(self, data: pd.DataFrame, by: str) -> pd.DataFrame:
        return data.sort_values(by)

    def filter_ats_on_ids(self, ats: pd.DataFrame, ids: List[str]) -> pd.DataFrame:
        return ats[ats['CitizenId'].isin(ids)]

    def remove_tainted_histories(self, ats: pd.DataFrame) -> pd.DataFrame:
        tained_ids = ats[ats['DevHMINumber'] == '899,999']['CitizenId'].unique()
        ats = ats[np.logical_not(ats['CitizenId'].isin(tained_ids))]
        return ats

    def remove_deprecated_device_data(self, ats: pd.DataFrame) -> pd.DataFrame:
        return ats[ats['DevHMINumber'] != '899,999']

    def remove_rows_with_old_dates(self, ats: pd.DataFrame, col: str) -> pd.DataFrame:
        ats[col] = pd.to_datetime(ats[col])
        mask = (ats[col] >= '1900-01-01') & (ats[col] <= pd.Timestamp('today'))
        return ats.loc[mask]
    
    def drop_invalid_devices(self, ats: pd.DataFrame, iso_classes: pd.DataFrame) -> pd.DataFrame:
        return ats[ats['DevISOClass'].isin(iso_classes.DevISOClass)]

    def remove_screenings_without_exercises(self, df: pd.DataFrame) -> pd.DataFrame:    
        df = df[df['ExerciseContent'] != 'nan']
        return df

class Cleaner2020(BaseCleaner):
    def clean_clusters(self, cl: pd.DataFrame) -> pd.DataFrame:
        return cl
    
    def clean_patient_data(self, ptd: pd.DataFrame) -> pd.DataFrame:
        ptd = self.remove_citizens_without_valid_id(ptd)
        return ptd
    
    def clean_screening_content(self, sc: pd.DataFrame, ptd: pd.DataFrame) -> pd.DataFrame:
        sc = self.remove_citizens_not_in_patient_data(sc, ptd, 'CitizenId')
        sc = self.merge_train_and_patient_data(sc, ptd, 'CitizenId')
        sc = self.sort_dataframe(sc, ['CitizenId', 'ScreeningDate'])
        return sc
    
    def clean_status_set(self, ss: pd.DataFrame, ptd: pd.DataFrame) -> pd.DataFrame:
        ss = self.remove_citizens_not_in_patient_data(ss, ptd, 'CitizenId')
        ss = self.merge_train_and_patient_data(ss, ptd, 'CitizenId')
        ss = self.sort_dataframe(ss, ['CitizenId', 'ChangeDate'])
        return ss
        
    def clean_training_done(self, td: pd.DataFrame, ptd: pd.DataFrame) -> pd.DataFrame:
        td = self.remove_citizens_not_in_patient_data(td, ptd, 'CitizenId')
        td = self.sort_dataframe(td, ['CitizenId', 'RatingDate'])
        return td

    def clean_training_cancelled(self, tc: pd.DataFrame, ptd: pd.DataFrame) -> pd.DataFrame:
        tc = self.remove_citizens_not_in_patient_data(tc, ptd, 'CitizenId')
        tc = self.merge_train_and_patient_data(tc, ptd, 'CitizenId')
        tc = self.sort_dataframe(tc, ['CitizenId', 'RatingDate'])
        return tc
    
    def clean_assistive_aids(self, ats: pd.DataFrame, ic: pd.DataFrame,
                             ids: List[str]=None) -> pd.DataFrame:
        ats = self.sort_dataframe(ats, ['CitizenId', 'LendDate'])
        ats = self.remove_citizens_without_valid_id(ats)
        ats = self.remove_rows_with_old_dates(ats, 'LendDate')
        ats = self.remove_deprecated_device_data(ats)
        ats = self.remove_tainted_histories(ats)
        ats = self.drop_invalid_devices(ats, ic)
        return ats
    
    def clean_fall_data(self, fd: pd.DataFrame) -> pd.DataFrame:
        fd = self.remove_citizens_without_valid_id(fd)
        fd = self.sort_dataframe(fd, ['CitizenId', 'Date'])
        return fd