import numpy as np
import pandas as pd
from tensorflow.python.data.ops.dataset_ops import normalize_to_dense
import paths as pt
from abc import ABC, abstractmethod
from tools import file_reader, preprocessor
from sklearn.preprocessing import StandardScaler, MinMaxScaler
from sklearn.model_selection import train_test_split
from scipy.stats import skew, boxcox

class BaseDataLoader(ABC):
    def __init__(self, file_name, settings, converters=None):
        """Initilizer method that takes a file name, settings and optionally a converter"""
        self.file_name = file_name
        self.settings = settings
        self.converters = converters
        
    @abstractmethod
    def load_data(self):
        """Loads the data from a data set at startup"""
        
    def get_data(self):
        """Returns the features and target"""
        return self.X, self.y
    
    def get_features(self):
        """Returns the feature names"""
        return self.X.columns
        
    def prepare_data(self):
        """Prepares the data from a data set"""
        X = self.X
        y = self.y
        
        features_to_normalize = self.settings['features_to_normalize']
        features_to_scale = self.settings['features_to_scale']
        X = preprocessor.normalize_data(X, features_to_normalize)
        X = preprocessor.scale_data(X, features_to_scale)
        
        X = np.array(X)
        y = np.array(y)
        return X, y
        
    def prepare_data_split(self, test_size):
        """Prepares and splits the data from a data set"""
        X = self.X
        y = self.y
        
        features_to_normalize = self.settings['features_to_normalize']
        features_to_scale = self.settings['features_to_scale']
        X = preprocessor.normalize_data(X, features_to_normalize)
        X = preprocessor.scale_data(X, features_to_scale)
        
        X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=test_size,
                                                            stratify=y, random_state=0)
        
        X_train = np.array(X_train)
        X_test = np.array(X_test)
        y_train = np.array(y_train)
        y_test = np.array(y_test)
    
        return X_train, X_test, y_train, y_test

class CompleteDataLoader(BaseDataLoader):
    def load_data(self):
        df = file_reader.read_csv(pt.PROCESSED_DATA_DIR,
                                  self.file_name,
                                  converters=self.converters)
        X = df.drop(['Complete'], axis=1)
        y = df['Complete']
        self.X = X
        self.y = y
        return self
    
class ComplianceDataLoader(BaseDataLoader):
    def load_data(self):
        df = file_reader.read_csv(pt.PROCESSED_DATA_DIR,
                                  self.file_name,
                                  converters=self.converters)
        X = df.drop(['Compliance'], axis=1)
        y = df['Compliance']
        self.X = X
        self.y = y
        return self

class FallDataLoader(BaseDataLoader):
    def load_data(self):
        df = file_reader.read_csv(pt.PROCESSED_DATA_DIR,
                                  self.file_name,
                                  converters=self.converters)
        X = df.drop(['Fall'], axis=1)
        y = df['Fall']
        self.X = X
        self.y = y
        return self

class RiskDataLoader(BaseDataLoader):
    def load_data(self):
        df = file_reader.read_csv(pt.PROCESSED_DATA_DIR,
                                  self.file_name,
                                  converters=self.converters)
        X = df.drop(['Risk'], axis=1)
        y = df['Risk']
        self.X = X
        self.y = y
        return self