import numpy as np
import pandas as pd
import paths as pt
from sklearn.model_selection import GridSearchCV
from xgboost import XGBClassifier
from tools import data_loader, file_writer
from sklearn.model_selection import StratifiedKFold
from sklearn.ensemble import RandomForestClassifier
from sklearn.metrics import roc_auc_score, accuracy_score
from imblearn.pipeline import make_pipeline, Pipeline
from imblearn.over_sampling import ADASYN
import matplotlib.pyplot as plt
from pathlib import Path
import paths as pt
import yaml
from sklearn.model_selection import cross_val_score

def main():
    with open(Path.joinpath(pt.CONFIGS_DIR, "settings.yaml"), 'r') as stream:
        settings = yaml.safe_load(stream)
    
    dl = data_loader.CompleteDataLoader("complete_emb.csv", settings).load_data()
    X, y = dl.get_data()
            
    param_grid = { 
        'model__n_estimators': [200, 400, 600, 800, 1000],
        'model__max_features': ['auto'],
        'model__max_depth' : [3],
        'model__min_samples_split' : [2],
        'model__min_samples_leaf': [3],
        'model__criterion' :['gini']
    }
        
    rf = RandomForestClassifier(random_state=0)
    adasyn = ADASYN(random_state=0)
    pipeline = Pipeline([('sampling', adasyn), ('model', rf)])
    
    rounds = 20
    outer_scores = np.zeros(rounds)
    nested_scores =  (rounds)
    metric = 'average_precision'
    
    for i in range(rounds):
        #Define both cross-validation objects (inner & outer)
        inner_cv = StratifiedKFold(n_splits=5, shuffle=True, random_state=i)
        outer_cv = StratifiedKFold(n_splits=5, shuffle=True, random_state=i)

        # Non-nested parameter search and scoring
        clf = GridSearchCV(estimator=pipeline, param_grid=param_grid,
                           cv=inner_cv, scoring=metric)
        clf.fit(X, y)
        outer_scores[i] = clf.best_score_

        # Nested CV with parameter optimization
        nested_score = cross_val_score(clf, X=X, y=y, cv=outer_cv,
                                       scoring=metric)
        nested_scores[i] = nested_score.mean()
    
    score_difference = outer_scores - nested_scores
    print("Avg. difference of {:6f} with std. dev. of {:6f}."
          .format(score_difference.mean(), score_difference.std()))
    
    # Plot scores on each round for nested and non-nested cross-validation
    plt.style.use('seaborn')
    plt.tight_layout()
    plt.figure(figsize=(10,5))
    outer_scores_line, = plt.plot(outer_scores, color='orange')
    nested_line, = plt.plot(nested_scores, color='steelblue')
    plt.ylabel("Score", fontsize="14")
    plt.legend([outer_scores_line, nested_line],
            ["Non-Nested CV", "Nested CV"],
            bbox_to_anchor=(0, .4, .5, 0))
    plt.title("Non-Nested vs Nested Cross-Validation",
            x=.5, y=1.1, fontsize="15")
    
    # Plot bar chart of the difference.
    plt.figure(figsize=(10,5))
    plt.tight_layout()
    difference_plot = plt.bar(range(rounds), score_difference)
    plt.xlabel("Individual Trial #")
    plt.legend([difference_plot],
            ["Non-Nested CV - Nested CV Score"],
            bbox_to_anchor=(0, 1, .8, 0))
    plt.ylabel("score difference", fontsize="14")
    plt.show()
        
if __name__ == "__main__":
    main()